from __future__ import annotations

import tkinter as tk

from .constants import (  # noqa: F401
    ICON_ADD,
    ICON_CLEAR,
    ICON_COPY,
    ICON_CUT,
    ICON_DEL,
    ICON_EDIT,
    ICON_PASTE,
    ICON_REDO,
    ICON_SELECT_ALL,
    ICON_SORT_ASC,
    ICON_SORT_DESC,
    ICON_UNDO,
    USER_OS,
    ctrl_key,
)
from .other_classes import DotDict, FontTuple
from .sorting import fast_sort_key, natural_sort_key, version_sort_key  # noqa: F401
from .themes import theme_light_blue


def new_sheet_options() -> DotDict:
    return DotDict(
        {
            **theme_light_blue,
            "popup_menu_font": FontTuple(
                "Calibri",
                13 if USER_OS == "darwin" else 11,
                "normal",
            ),
            "table_font": FontTuple(
                "Calibri",
                13 if USER_OS == "darwin" else 11,
                "normal",
            ),
            "header_font": FontTuple(
                "Calibri",
                13 if USER_OS == "darwin" else 11,
                "normal",
            ),
            "index_font": FontTuple(
                "Calibri",
                13 if USER_OS == "darwin" else 11,
                "normal",
            ),
            # edit header
            "edit_header_label": "Edit header",
            "edit_header_accelerator": "",
            "edit_header_image": tk.PhotoImage(data=ICON_EDIT),
            "edit_header_compound": "left",
            # edit index
            "edit_index_label": "Edit index",
            "edit_index_accelerator": "",
            "edit_index_image": tk.PhotoImage(data=ICON_EDIT),
            "edit_index_compound": "left",
            # edit cell
            "edit_cell_label": "Edit cell",
            "edit_cell_accelerator": "",
            "edit_cell_image": tk.PhotoImage(data=ICON_EDIT),
            "edit_cell_compound": "left",
            # cut
            "cut_label": "Cut",
            "cut_accelerator": "Ctrl+X",
            "cut_image": tk.PhotoImage(data=ICON_CUT),
            "cut_compound": "left",
            # cut contents
            "cut_contents_label": "Cut contents",
            "cut_contents_accelerator": "Ctrl+X",
            "cut_contents_image": tk.PhotoImage(data=ICON_CUT),
            "cut_contents_compound": "left",
            # copy
            "copy_label": "Copy",
            "copy_accelerator": "Ctrl+C",
            "copy_image": tk.PhotoImage(data=ICON_COPY),
            "copy_compound": "left",
            # copy contents
            "copy_contents_label": "Copy contents",
            "copy_contents_accelerator": "Ctrl+C",
            "copy_contents_image": tk.PhotoImage(data=ICON_COPY),
            "copy_contents_compound": "left",
            # paste
            "paste_label": "Paste",
            "paste_accelerator": "Ctrl+V",
            "paste_image": tk.PhotoImage(data=ICON_PASTE),
            "paste_compound": "left",
            # delete
            "delete_label": "Delete",
            "delete_accelerator": "Del",
            "delete_image": tk.PhotoImage(data=ICON_CLEAR),
            "delete_compound": "left",
            # clear contents
            "clear_contents_label": "Clear contents",
            "clear_contents_accelerator": "Del",
            "clear_contents_image": tk.PhotoImage(data=ICON_CLEAR),
            "clear_contents_compound": "left",
            # del columns
            "delete_columns_label": "Delete columns",
            "delete_columns_accelerator": "",
            "delete_columns_image": tk.PhotoImage(data=ICON_DEL),
            "delete_columns_compound": "left",
            # insert columns left
            "insert_columns_left_label": "Insert columns left",
            "insert_columns_left_accelerator": "",
            "insert_columns_left_image": tk.PhotoImage(data=ICON_ADD),
            "insert_columns_left_compound": "left",
            # insert columns right
            "insert_columns_right_label": "Insert columns right",
            "insert_columns_right_accelerator": "",
            "insert_columns_right_image": tk.PhotoImage(data=ICON_ADD),
            "insert_columns_right_compound": "left",
            # insert single column
            "insert_column_label": "Insert column",
            "insert_column_accelerator": "",
            "insert_column_image": tk.PhotoImage(data=ICON_ADD),
            "insert_column_compound": "left",
            # del rows
            "delete_rows_label": "Delete rows",
            "delete_rows_accelerator": "",
            "delete_rows_image": tk.PhotoImage(data=ICON_DEL),
            "delete_rows_compound": "left",
            # insert rows above
            "insert_rows_above_label": "Insert rows above",
            "insert_rows_above_accelerator": "",
            "insert_rows_above_image": tk.PhotoImage(data=ICON_ADD),
            "insert_rows_above_compound": "left",
            # insert rows below
            "insert_rows_below_label": "Insert rows below",
            "insert_rows_below_accelerator": "",
            "insert_rows_below_image": tk.PhotoImage(data=ICON_ADD),
            "insert_rows_below_compound": "left",
            # insert single row
            "insert_row_label": "Insert row",
            "insert_row_accelerator": "",
            "insert_row_image": tk.PhotoImage(data=ICON_ADD),
            "insert_row_compound": "left",
            # sorting
            # labels
            "sort_cells_label": "Sort Asc.",
            "sort_cells_x_label": "Sort row-wise Asc.",
            "sort_row_label": "Sort values Asc.",
            "sort_column_label": "Sort values Asc.",
            "sort_rows_label": "Sort rows Asc.",
            "sort_columns_label": "Sort columns Asc.",
            # reverse labels
            "sort_cells_reverse_label": "Sort Desc.",
            "sort_cells_x_reverse_label": "Sort row-wise Desc.",
            "sort_row_reverse_label": "Sort values Desc.",
            "sort_column_reverse_label": "Sort values Desc.",
            "sort_rows_reverse_label": "Sort rows Desc.",
            "sort_columns_reverse_label": "Sort columns Desc.",
            # accelerators
            "sort_cells_accelerator": "",
            "sort_cells_x_accelerator": "",
            "sort_row_accelerator": "",
            "sort_column_accelerator": "",
            "sort_rows_accelerator": "",
            "sort_columns_accelerator": "",
            # reverse accelerators
            "sort_cells_reverse_accelerator": "",
            "sort_cells_x_reverse_accelerator": "",
            "sort_row_reverse_accelerator": "",
            "sort_column_reverse_accelerator": "",
            "sort_rows_reverse_accelerator": "",
            "sort_columns_reverse_accelerator": "",
            # images
            "sort_cells_image": tk.PhotoImage(data=ICON_SORT_ASC),
            "sort_cells_x_image": tk.PhotoImage(data=ICON_SORT_ASC),
            "sort_row_image": tk.PhotoImage(data=ICON_SORT_ASC),
            "sort_column_image": tk.PhotoImage(data=ICON_SORT_ASC),
            "sort_rows_image": tk.PhotoImage(data=ICON_SORT_ASC),
            "sort_columns_image": tk.PhotoImage(data=ICON_SORT_ASC),
            # compounds
            "sort_cells_compound": "left",
            "sort_cells_x_compound": "left",
            "sort_row_compound": "left",
            "sort_column_compound": "left",
            "sort_rows_compound": "left",
            "sort_columns_compound": "left",
            # reverse images
            "sort_cells_reverse_image": tk.PhotoImage(data=ICON_SORT_DESC),
            "sort_cells_x_reverse_image": tk.PhotoImage(data=ICON_SORT_DESC),
            "sort_row_reverse_image": tk.PhotoImage(data=ICON_SORT_DESC),
            "sort_column_reverse_image": tk.PhotoImage(data=ICON_SORT_DESC),
            "sort_rows_reverse_image": tk.PhotoImage(data=ICON_SORT_DESC),
            "sort_columns_reverse_image": tk.PhotoImage(data=ICON_SORT_DESC),
            # reverse compounds
            "sort_cells_reverse_compound": "left",
            "sort_cells_x_reverse_compound": "left",
            "sort_row_reverse_compound": "left",
            "sort_column_reverse_compound": "left",
            "sort_rows_reverse_compound": "left",
            "sort_columns_reverse_compound": "left",
            # select all
            "select_all_label": "Select all",
            "select_all_accelerator": "Ctrl+A",
            "select_all_image": tk.PhotoImage(data=ICON_SELECT_ALL),
            "select_all_compound": "left",
            # undo
            "undo_label": "Undo",
            "undo_accelerator": "Ctrl+Z",
            "undo_image": tk.PhotoImage(data=ICON_UNDO),
            "undo_compound": "left",
            # redo
            "redo_label": "Redo",
            "redo_accelerator": "Ctrl+Shift+Z",
            "redo_image": tk.PhotoImage(data=ICON_REDO),
            "redo_compound": "left",
            # bindings
            "copy_bindings": [
                f"<{ctrl_key}-c>",
                f"<{ctrl_key}-C>",
            ],
            "cut_bindings": [
                f"<{ctrl_key}-x>",
                f"<{ctrl_key}-X>",
            ],
            "paste_bindings": [
                f"<{ctrl_key}-v>",
                f"<{ctrl_key}-V>",
            ],
            "undo_bindings": [
                f"<{ctrl_key}-z>",
                f"<{ctrl_key}-Z>",
            ],
            "redo_bindings": [
                f"<{ctrl_key}-Shift-z>",
                f"<{ctrl_key}-Shift-Z>",
            ],
            "delete_bindings": [
                "<Delete>",
            ],
            "select_all_bindings": [
                f"<{ctrl_key}-a>",
                f"<{ctrl_key}-A>",
                f"<{ctrl_key}-Shift-space>",
            ],
            "select_columns_bindings": [
                "<Control-space>",
            ],
            "select_rows_bindings": [
                "<Shift-space>",
            ],
            "row_start_bindings": [
                "<Command-Left>",
                "<Home>",
            ]
            if USER_OS == "darwin"
            else ["<Home>"],
            "table_start_bindings": [
                f"<{ctrl_key}-Home>",
            ],
            "tab_bindings": [
                "<Tab>",
            ],
            "up_bindings": [
                "<Up>",
            ],
            "right_bindings": [
                "<Right>",
            ],
            "down_bindings": [
                "<Down>",
            ],
            "left_bindings": [
                "<Left>",
            ],
            "shift_up_bindings": [
                "<Shift-Up>",
            ],
            "shift_right_bindings": [
                "<Shift-Right>",
            ],
            "shift_down_bindings": [
                "<Shift-Down>",
            ],
            "shift_left_bindings": [
                "<Shift-Left>",
            ],
            "prior_bindings": [
                "<Prior>",
            ],
            "next_bindings": [
                "<Next>",
            ],
            "find_bindings": [
                f"<{ctrl_key}-f>",
                f"<{ctrl_key}-F>",
            ],
            "find_next_bindings": [
                f"<{ctrl_key}-g>",
                f"<{ctrl_key}-G>",
            ],
            "find_previous_bindings": [
                f"<{ctrl_key}-Shift-g>",
                f"<{ctrl_key}-Shift-G>",
            ],
            "toggle_replace_bindings": [
                f"<{ctrl_key}-h>",
                f"<{ctrl_key}-H>",
            ],
            "escape_bindings": [
                "<Escape>",
            ],
            # other
            "vertical_scroll_borderwidth": 1,
            "horizontal_scroll_borderwidth": 1,
            "vertical_scroll_gripcount": 0,
            "horizontal_scroll_gripcount": 0,
            "vertical_scroll_arrowsize": "",
            "horizontal_scroll_arrowsize": "",
            "set_cell_sizes_on_zoom": False,
            "auto_resize_columns": None,
            "auto_resize_rows": None,
            "to_clipboard_delimiter": "\t",
            "to_clipboard_quotechar": '"',
            "to_clipboard_lineterminator": "\n",
            "from_clipboard_delimiters": ["\t"],
            "show_dropdown_borders": False,
            "show_default_header_for_empty": True,
            "show_default_index_for_empty": True,
            "default_header_height": "1",
            "default_row_height": "1",
            "default_column_width": 120,
            "default_row_index_width": 70,
            "default_row_index": "numbers",
            "default_header": "letters",
            "page_up_down_select_row": True,
            "paste_can_expand_x": False,
            "paste_can_expand_y": False,
            "paste_insert_column_limit": None,
            "paste_insert_row_limit": None,
            "arrow_key_down_right_scroll_page": False,
            "cell_auto_resize_enabled": True,
            "auto_resize_row_index": True,
            "max_undos": 30,
            "column_drag_and_drop_perform": True,
            "row_drag_and_drop_perform": True,
            "empty_horizontal": 50,
            "empty_vertical": 50,
            "horizontal_grid_to_end_of_window": False,
            "vertical_grid_to_end_of_window": False,
            "show_vertical_grid": True,
            "show_horizontal_grid": True,
            "display_selected_fg_over_highlights": False,
            "show_selected_cells_border": True,
            "edit_cell_tab": "right",
            "edit_cell_return": "down",
            "editor_del_key": "forward",
            "treeview": False,
            "treeview_indent": "2",
            "rounded_boxes": True,
            "alternate_color": "",
            "allow_cell_overflow": False,
            "table_wrap": "c",
            "header_wrap": "c",
            "index_wrap": "c",
            "min_column_width": 1,
            "max_column_width": float("inf"),
            "max_header_height": float("inf"),
            "max_row_height": float("inf"),
            "max_index_width": float("inf"),
            "show_top_left": None,
            "sort_key": natural_sort_key,
            "tooltips": False,
            "user_can_create_notes": False,
            "note_corners": False,
            "tooltip_width": 210,
            "tooltip_height": 210,
        }
    )



from __future__ import annotations

import tkinter as tk
from tkinter import ttk
from typing import Any, Literal

from .constants import ctrl_key, rc_binding


class TooltipTkText(tk.Text):
    """Custom Text widget for the Tooltip class."""

    def __init__(
        self,
        parent: tk.Misc,
    ) -> None:
        super().__init__(
            parent,
            spacing1=0,
            spacing2=1,
            spacing3=0,
            bd=0,
            highlightthickness=0,
            undo=True,
            maxundo=30,
        )
        self.parent = parent
        self.rc_popup_menu = tk.Menu(self, tearoff=0)
        self.bind("<1>", lambda event: self.focus_set())
        self.bind(rc_binding, self.rc)
        self.bind(f"<{ctrl_key}-a>", self.select_all)
        self.bind(f"<{ctrl_key}-A>", self.select_all)
        self.bind("<Delete>", self.delete_key)

    def reset(
        self,
        menu_kwargs: dict,
        sheet_ops: dict,
        bg: str,
        fg: str,
        select_bg: str,
        select_fg: str,
        text: str,
        readonly: bool,
    ) -> None:
        """Reset the text widget's appearance and menu options."""
        self.config(
            font=sheet_ops.table_font,
            background=bg,
            foreground=fg,
            insertbackground=fg,
            selectbackground=select_bg,
            selectforeground=select_fg,
        )
        self.editor_del_key = sheet_ops.editor_del_key
        self.config(state="normal")
        self.delete(1.0, "end")
        self.insert(1.0, text)
        self.edit_reset()
        self.rc_popup_menu.delete(0, "end")
        self.rc_popup_menu.add_command(
            label=sheet_ops.select_all_label,
            accelerator=sheet_ops.select_all_accelerator,
            image=sheet_ops.select_all_image,
            compound=sheet_ops.select_all_compound,
            command=self.select_all,
            **menu_kwargs,
        )
        if not readonly:
            self.rc_popup_menu.add_command(
                label=sheet_ops.cut_label,
                accelerator=sheet_ops.cut_accelerator,
                image=sheet_ops.cut_image,
                compound=sheet_ops.cut_compound,
                command=self.cut,
                **menu_kwargs,
            )
        self.rc_popup_menu.add_command(
            label=sheet_ops.copy_label,
            accelerator=sheet_ops.copy_accelerator,
            image=sheet_ops.copy_image,
            compound=sheet_ops.copy_compound,
            command=self.copy,
            **menu_kwargs,
        )
        if not readonly:
            self.rc_popup_menu.add_command(
                label=sheet_ops.paste_label,
                accelerator=sheet_ops.paste_accelerator,
                image=sheet_ops.paste_image,
                compound=sheet_ops.paste_compound,
                command=self.paste,
                **menu_kwargs,
            )
            self.rc_popup_menu.add_command(
                label=sheet_ops.undo_label,
                accelerator=sheet_ops.undo_accelerator,
                image=sheet_ops.undo_image,
                compound=sheet_ops.undo_compound,
                command=self.undo,
                **menu_kwargs,
            )
            self.rc_popup_menu.add_command(
                label=sheet_ops.redo_label,
                accelerator=sheet_ops.redo_accelerator,
                image=sheet_ops.redo_image,
                compound=sheet_ops.redo_compound,
                command=self.redo,
                **menu_kwargs,
            )

    def rc(self, event: Any) -> None:
        """Show the right-click popup menu."""
        self.rc_popup_menu.tk_popup(event.x_root, event.y_root)
        self.focus_set()

    def delete_key(self, event: Any = None) -> None:
        """Handle the Delete key based on editor configuration."""
        if self.editor_del_key == "forward":
            return
        elif not self.editor_del_key:
            return "break"
        elif self.editor_del_key == "backward":
            if self.tag_ranges("sel"):
                return
            if self.index("insert") == "1.0":
                return "break"
            self.delete("insert-1c")
            return "break"

    def select_all(self, event: Any = None) -> Literal["break"]:
        """Select all text in the widget."""
        self.tag_add(tk.SEL, "1.0", "end-1c")
        self.mark_set(tk.INSERT, "end-1c")
        return "break"

    def cut(self, event: Any = None) -> Literal["break"]:
        """Cut selected text."""
        self.event_generate(f"<{ctrl_key}-x>")
        self.event_generate("<KeyRelease>")
        return "break"

    def copy(self, event: Any = None) -> Literal["break"]:
        """Copy selected text."""
        self.event_generate(f"<{ctrl_key}-c>")
        return "break"

    def paste(self, event: Any = None) -> Literal["break"]:
        """Paste text from clipboard."""
        self.event_generate(f"<{ctrl_key}-v>")
        self.event_generate("<KeyRelease>")
        return "break"

    def undo(self, event: Any = None) -> Literal["break"]:
        """Undo the last action."""
        self.event_generate(f"<{ctrl_key}-z>")
        self.event_generate("<KeyRelease>")
        return "break"

    def redo(self, event: Any = None) -> Literal["break"]:
        self.event_generate(f"<{ctrl_key}-Shift-z>")
        self.event_generate("<KeyRelease>")
        return "break"


class Tooltip(tk.Toplevel):
    def __init__(
        self,
        parent: tk.Misc,
        sheet_ops: dict,
        menu_kwargs: dict,
        bg: str,
        fg: str,
        select_bg: str,
        select_fg: str,
        scrollbar_style: str,
    ) -> None:
        super().__init__(parent)
        self.withdraw()  # Hide until positioned
        self.overrideredirect(True)  # Borderless window
        self.cell_readonly = True
        self.note_readonly = True

        # Store parameters
        self.sheet_ops = sheet_ops
        self.menu_kwargs = menu_kwargs
        self.font = sheet_ops.table_font
        self.bg = bg
        self.fg = fg
        self.select_bg = select_bg
        self.select_fg = select_fg
        self.row = 0
        self.col = 0

        # Create border frame for visual distinction
        self.border_frame = tk.Frame(self, background=bg)
        self.border_frame.pack(fill="both", expand=True, padx=1, pady=1)

        # Create notebook, but don’t pack it yet
        self.notebook = ttk.Notebook(self.border_frame)

        # Content frame as child of border_frame
        self.content_frame = ttk.Frame(self.border_frame)
        self.content_text = TooltipTkText(self.content_frame)
        self.content_scrollbar = ttk.Scrollbar(self.content_frame, orient="vertical", style=scrollbar_style)
        self.content_scrollbar.pack(side="right", fill="y")
        self.content_text.pack(side="left", fill="both", expand=True)
        self.content_scrollbar.configure(command=self.content_text.yview)
        self.content_text.configure(yscrollcommand=self.content_scrollbar.set)

        # Note frame as child of border_frame
        self.note_frame = ttk.Frame(self.border_frame)
        self.note_text = TooltipTkText(self.note_frame)
        self.note_scrollbar = ttk.Scrollbar(self.note_frame, orient="vertical", style=scrollbar_style)
        self.note_scrollbar.pack(side="right", fill="y")
        self.note_text.pack(side="left", fill="both", expand=True)
        self.note_scrollbar.configure(command=self.note_text.yview)
        self.note_text.configure(yscrollcommand=self.note_scrollbar.set)

    def setup_note_only_mode(self):
        """Configure the tooltip to show only the note text widget."""
        self.notebook.pack_forget()  # Remove notebook from layout
        for tab in self.notebook.tabs():  # Clear all tabs to free the frames
            self.notebook.forget(tab)
        self.content_frame.pack_forget()  # Ensure content_frame is not directly packed
        self.note_frame.pack(fill="both", expand=True)  # Show note_frame directly

    def setup_single_text_mode(self):
        """Configure the tooltip to show only the content text widget."""
        self.notebook.pack_forget()  # Remove notebook from layout
        for tab in self.notebook.tabs():  # Clear all tabs
            self.notebook.forget(tab)
        self.note_frame.pack_forget()  # Ensure note_frame is not directly packed
        self.content_frame.pack(fill="both", expand=True)  # Show content_frame directly

    def setup_notebook_mode(self):
        """Configure the tooltip to show a notebook with Cell and Note tabs."""
        self.content_frame.pack_forget()  # Ensure content_frame is not directly packed
        self.note_frame.pack_forget()  # Ensure note_frame is not directly packed
        self.notebook.add(self.content_frame, text="Cell")  # Add content tab
        self.notebook.add(self.note_frame, text="Note")  # Add note tab
        self.notebook.pack(fill="both", expand=True)  # Show notebook
        self.notebook.select(0)  # Select the "Cell" tab by default

    def reset(
        self,
        text: str,
        cell_readonly: bool,
        note: str | None,
        note_readonly: bool,
        row: int,
        col: int,
        menu_kwargs: dict,
        bg: str,
        fg: str,
        select_bg: str,
        select_fg: str,
        user_can_create_notes: bool,
        note_only: bool,
        width: int,
        height: int,
    ) -> None:
        self.cell_readonly = cell_readonly
        self.note_readonly = note_readonly
        self.menu_kwargs = menu_kwargs
        self.font = self.sheet_ops.table_font
        self.bg = bg
        self.fg = fg
        self.select_bg = select_bg
        self.select_fg = select_fg
        self.row = row
        self.col = col
        self.config(bg=self.sheet_ops.table_selected_box_cells_fg)
        self.border_frame.config(background=bg)
        self.content_text.configure(
            wrap="word",
            width=30,
            height=5,
            state="disabled" if cell_readonly else "normal",
        )
        kws = {
            "menu_kwargs": menu_kwargs,
            "sheet_ops": self.sheet_ops,
            "bg": bg,
            "fg": fg,
            "select_bg": select_bg,
            "select_fg": select_fg,
        }
        self.content_text.reset(**kws, text=text, readonly=cell_readonly)
        self.content_text.configure(state="disabled" if cell_readonly else "normal")
        self.note_text.configure(
            wrap="word",
            width=30,
            height=5,
        )
        self.note_text.reset(**kws, text="" if note is None else note, readonly=note_readonly)
        self.note_text.configure(state="disabled" if note_readonly else "normal")
        # Set up UI based on condition
        if note_only:
            self.setup_note_only_mode()
        elif note is None and not user_can_create_notes:
            self.setup_single_text_mode()
        else:
            self.setup_notebook_mode()
        self.adjust_size(width, height)

    def adjust_size(self, width, height) -> None:
        """Adjust tooltip size to given dimensions."""
        self.update_idletasks()
        self.geometry(f"{width}x{height}")

    def set_position(self, x_root: int, y_root: int) -> None:
        """Position tooltip near the cell, avoiding screen edges."""
        self.update_idletasks()
        screen_width = self.winfo_screenwidth()
        screen_height = self.winfo_screenheight()
        width = self.winfo_width()
        height = self.winfo_height()

        # Adjust position to avoid screen edges
        x = min(x_root, screen_width - width)
        y = min(y_root, screen_height - height)
        self.geometry(f"+{x}+{y}")
        self.deiconify()

    def get(self) -> tuple[int, int, str, str]:
        return (
            self.row,
            self.col,
            self.content_text.get("1.0", "end-1c"),
            self.note_text.get("1.0", "end-1c"),
        )

    def destroy(self) -> None:
        super().destroy()



from __future__ import annotations

from platform import system as get_os
from typing import Hashable

USER_OS: str = f"{get_os()}".lower()
ctrl_key: str = "Command" if USER_OS == "darwin" else "Control"
rc_binding: str = "<2>" if USER_OS == "darwin" else "<3>"
symbols_set: set[str] = set("""!#$%&'()*+,-./:;"@[]^_`{|}~>?= \\""")
nonelike: set[Hashable] = {None, "none", ""}
truthy: set[Hashable] = {True, "true", "t", "yes", "y", "on", "1"}
falsy: set[Hashable] = {False, "false", "f", "no", "n", "off", "0"}
_test_str: str = "aiW_-|"

val_modifying_options: set[str] = {"checkbox", "format", "dropdown"}

named_span_types: set[str] = {
    "format",
    "highlight",
    "dropdown",
    "checkbox",
    "readonly",
    "align",
    "note",
}

emitted_events: set[str] = {
    "<<SheetModified>>",
    "<<SheetRedrawn>>",
    "<<SheetSelect>>",
    "<<Copy>>",
    "<<Cut>>",
    "<<Paste>>",
    "<<Delete>>",
    "<<Undo>>",
    "<<Redo>>",
    "<<SelectAll>>",
}

align_helper: dict[str, str] = {
    "w": "nw",
    "e": "ne",
    "center": "n",
    "n": "center",
    "nw": "left",
    "ne": "right",
}

align_value_error: str = "Align must be one of the following values: c, center, w, nw, west, left, e, ne, east, right"
font_value_error: str = "Argument must be font, size and 'normal', 'bold' or'italic' e.g. ('Carlito',12,'normal')"

backwards_compatibility_keys: dict[str, str] = {
    "font": "table_font",
}

text_editor_newline_bindings: set[str] = {
    "<Alt-Return>",
    "<Alt-KP_Enter>",
    "<Shift-Return>",
    f"<{ctrl_key}-Return>",
}
if USER_OS == "darwin":
    text_editor_newline_bindings.add("<Option-Return>")

text_editor_close_bindings: dict[str, str] = {
    "<Tab>": "Tab",
    "<Return>": "Return",
    "<KP_Enter>": "Return",
    "<Escape>": "Escape",
}

text_editor_to_unbind: set[str] = (
    text_editor_newline_bindings | set(text_editor_close_bindings) | {"<FocusOut>", "<KeyRelease>"}
)

scrollbar_options_keys: set[str] = {
    "vertical_scroll_background",
    "horizontal_scroll_background",
    "vertical_scroll_troughcolor",
    "horizontal_scroll_troughcolor",
    "vertical_scroll_lightcolor",
    "horizontal_scroll_lightcolor",
    "vertical_scroll_darkcolor",
    "horizontal_scroll_darkcolor",
    "vertical_scroll_relief",
    "horizontal_scroll_relief",
    "vertical_scroll_troughrelief",
    "horizontal_scroll_troughrelief",
    "vertical_scroll_bordercolor",
    "horizontal_scroll_bordercolor",
    "vertical_scroll_borderwidth",
    "horizontal_scroll_borderwidth",
    "vertical_scroll_gripcount",
    "horizontal_scroll_gripcount",
    "vertical_scroll_arrowsize",
    "horizontal_scroll_arrowsize",
    "vertical_scroll_active_bg",
    "horizontal_scroll_active_bg",
    "vertical_scroll_not_active_bg",
    "horizontal_scroll_not_active_bg",
    "vertical_scroll_pressed_bg",
    "horizontal_scroll_pressed_bg",
    "vertical_scroll_active_fg",
    "horizontal_scroll_active_fg",
    "vertical_scroll_not_active_fg",
    "horizontal_scroll_not_active_fg",
    "vertical_scroll_pressed_fg",
    "horizontal_scroll_pressed_fg",
}

bind_add_columns: set[str] = {
    "all",
    "rc_insert_column",
    "insert_column",
    "add_column",
    "insert_columns",
    "add_columns",
    "insert column",
    "add column",
}

bind_del_columns: set[str] = {
    "all",
    "rc_delete_column",
    "delete_column",
    "del_column",
    "delete_columns",
    "del_columns",
    "delete columns",
    "del columns",
}

bind_add_rows: set[str] = {
    "all",
    "rc_insert_row",
    "insert_row",
    "add_row",
    "insert_rows",
    "add_rows",
    "insert row",
    "add row",
}

bind_del_rows: set[str] = {
    "all",
    "rc_delete_row",
    "delete_row",
    "del_row",
    "delete_rows",
    "del_rows",
    "delete rows",
    "del rows",
}

BINDING_TO_ATTR = {
    "replace_all": ("MT", "extra_end_replace_all_func"),
    # begin sort cells
    "begin_sort_cells": ("MT", "extra_begin_sort_cells_func"),
    # end sort cells
    "sort_cells": ("MT", "extra_end_sort_cells_func"),
    "end_sort_cells": ("MT", "extra_end_sort_cells_func"),
    # begin sort rows
    "begin_sort_rows": ("CH", "ch_extra_begin_sort_rows_func"),
    # end sort rows
    "sort_rows": ("CH", "ch_extra_end_sort_rows_func"),
    "end_sort_rows": ("CH", "ch_extra_end_sort_rows_func"),
    # begin sort cols
    "begin_sort_columns": ("RI", "ri_extra_begin_sort_cols_func"),
    # end sort cols
    "sort_columns": ("RI", "ri_extra_end_sort_cols_func"),
    "end_sort_columns": ("RI", "ri_extra_end_sort_cols_func"),
    # begin copy
    "begin_copy": ("MT", "extra_begin_ctrl_c_func"),
    "begin_ctrl_c": ("MT", "extra_begin_ctrl_c_func"),
    # end copy
    "ctrl_c": ("MT", "extra_end_ctrl_c_func"),
    "end_copy": ("MT", "extra_end_ctrl_c_func"),
    "end_ctrl_c": ("MT", "extra_end_ctrl_c_func"),
    "copy": ("MT", "extra_end_ctrl_c_func"),
    # begin cut
    "begin_cut": ("MT", "extra_begin_ctrl_x_func"),
    "begin_ctrl_x": ("MT", "extra_begin_ctrl_x_func"),
    # end cut
    "ctrl_x": ("MT", "extra_end_ctrl_x_func"),
    "end_cut": ("MT", "extra_end_ctrl_x_func"),
    "end_ctrl_x": ("MT", "extra_end_ctrl_x_func"),
    "cut": ("MT", "extra_end_ctrl_x_func"),
    # begin paste
    "begin_paste": ("MT", "extra_begin_ctrl_v_func"),
    "begin_ctrl_v": ("MT", "extra_begin_ctrl_v_func"),
    # end paste
    "ctrl_v": ("MT", "extra_end_ctrl_v_func"),
    "end_paste": ("MT", "extra_end_ctrl_v_func"),
    "end_ctrl_v": ("MT", "extra_end_ctrl_v_func"),
    "paste": ("MT", "extra_end_ctrl_v_func"),
    # begin undo/redo
    "begin_undo": ("MT", "extra_begin_ctrl_z_func"),
    "begin_ctrl_z": ("MT", "extra_begin_ctrl_z_func"),
    # end undo/redo
    "ctrl_z": ("MT", "extra_end_ctrl_z_func"),
    "end_undo": ("MT", "extra_end_ctrl_z_func"),
    "end_ctrl_z": ("MT", "extra_end_ctrl_z_func"),
    "undo": ("MT", "extra_end_ctrl_z_func"),
    # begin del key
    "begin_delete_key": ("MT", "extra_begin_delete_key_func"),
    "begin_delete": ("MT", "extra_begin_delete_key_func"),
    # end del key
    "delete_key": ("MT", "extra_end_delete_key_func"),
    "end_delete": ("MT", "extra_end_delete_key_func"),
    "end_delete_key": ("MT", "extra_end_delete_key_func"),
    "delete": ("MT", "extra_end_delete_key_func"),
    # begin edit table
    "begin_edit_cell": ("MT", "extra_begin_edit_cell_func"),
    "begin_edit_table": ("MT", "extra_begin_edit_cell_func"),
    # end edit table
    "end_edit_cell": ("MT", "extra_end_edit_cell_func"),
    "edit_cell": ("MT", "extra_end_edit_cell_func"),
    "edit_table": ("MT", "extra_end_edit_cell_func"),
    # begin edit header
    "begin_edit_header": ("CH", "extra_begin_edit_cell_func"),
    # end edit header
    "end_edit_header": ("CH", "extra_end_edit_cell_func"),
    "edit_header": ("CH", "extra_end_edit_cell_func"),
    # begin edit index
    "begin_edit_index": ("RI", "extra_begin_edit_cell_func"),
    # end edit index
    "end_edit_index": ("RI", "extra_end_edit_cell_func"),
    "edit_index": ("RI", "extra_end_edit_cell_func"),
    # begin move rows
    "begin_row_index_drag_drop": ("RI", "ri_extra_begin_drag_drop_func"),
    "begin_move_rows": ("RI", "ri_extra_begin_drag_drop_func"),
    # end move rows
    "row_index_drag_drop": ("RI", "ri_extra_end_drag_drop_func"),
    "move_rows": ("RI", "ri_extra_end_drag_drop_func"),
    "end_move_rows": ("RI", "ri_extra_end_drag_drop_func"),
    "end_row_index_drag_drop": ("RI", "ri_extra_end_drag_drop_func"),
    # begin move cols
    "begin_column_header_drag_drop": ("CH", "ch_extra_begin_drag_drop_func"),
    "begin_move_columns": ("CH", "ch_extra_begin_drag_drop_func"),
    # end move cols
    "column_header_drag_drop": ("CH", "ch_extra_end_drag_drop_func"),
    "move_columns": ("CH", "ch_extra_end_drag_drop_func"),
    "end_move_columns": ("CH", "ch_extra_end_drag_drop_func"),
    "end_column_header_drag_drop": ("CH", "ch_extra_end_drag_drop_func"),
    # begin del rows
    "begin_rc_delete_row": ("MT", "extra_begin_del_rows_rc_func"),
    "begin_delete_rows": ("MT", "extra_begin_del_rows_rc_func"),
    # end del rows
    "rc_delete_row": ("MT", "extra_end_del_rows_rc_func"),
    "end_rc_delete_row": ("MT", "extra_end_del_rows_rc_func"),
    "end_delete_rows": ("MT", "extra_end_del_rows_rc_func"),
    "delete_rows": ("MT", "extra_end_del_rows_rc_func"),
    # begin del cols
    "begin_rc_delete_column": ("MT", "extra_begin_del_cols_rc_func"),
    "begin_delete_columns": ("MT", "extra_begin_del_cols_rc_func"),
    # end del cols
    "rc_delete_column": ("MT", "extra_end_del_cols_rc_func"),
    "end_rc_delete_column": ("MT", "extra_end_del_cols_rc_func"),
    "end_delete_columns": ("MT", "extra_end_del_cols_rc_func"),
    "delete_columns": ("MT", "extra_end_del_cols_rc_func"),
    # begin add cols
    "begin_rc_insert_column": ("MT", "extra_begin_insert_cols_rc_func"),
    "begin_insert_column": ("MT", "extra_begin_insert_cols_rc_func"),
    "begin_insert_columns": ("MT", "extra_begin_insert_cols_rc_func"),
    "begin_add_column": ("MT", "extra_begin_insert_cols_rc_func"),
    "begin_rc_add_column": ("MT", "extra_begin_insert_cols_rc_func"),
    "begin_add_columns": ("MT", "extra_begin_insert_cols_rc_func"),
    # end add cols
    "rc_insert_column": ("MT", "extra_end_insert_cols_rc_func"),
    "end_rc_insert_column": ("MT", "extra_end_insert_cols_rc_func"),
    "end_insert_column": ("MT", "extra_end_insert_cols_rc_func"),
    "end_insert_columns": ("MT", "extra_end_insert_cols_rc_func"),
    "rc_add_column": ("MT", "extra_end_insert_cols_rc_func"),
    "end_rc_add_column": ("MT", "extra_end_insert_cols_rc_func"),
    "end_add_column": ("MT", "extra_end_insert_cols_rc_func"),
    "end_add_columns": ("MT", "extra_end_insert_cols_rc_func"),
    "add_columns": ("MT", "extra_end_insert_cols_rc_func"),
    # begin add rows
    "begin_rc_insert_row": ("MT", "extra_begin_insert_rows_rc_func"),
    "begin_insert_row": ("MT", "extra_begin_insert_rows_rc_func"),
    "begin_insert_rows": ("MT", "extra_begin_insert_rows_rc_func"),
    "begin_rc_add_row": ("MT", "extra_begin_insert_rows_rc_func"),
    "begin_add_row": ("MT", "extra_begin_insert_rows_rc_func"),
    "begin_add_rows": ("MT", "extra_begin_insert_rows_rc_func"),
    # end add rows
    "rc_insert_row": ("MT", "extra_end_insert_rows_rc_func"),
    "end_rc_insert_row": ("MT", "extra_end_insert_rows_rc_func"),
    "end_insert_row": ("MT", "extra_end_insert_rows_rc_func"),
    "end_insert_rows": ("MT", "extra_end_insert_rows_rc_func"),
    "rc_add_row": ("MT", "extra_end_insert_rows_rc_func"),
    "end_rc_add_row": ("MT", "extra_end_insert_rows_rc_func"),
    "end_add_row": ("MT", "extra_end_insert_rows_rc_func"),
    "end_add_rows": ("MT", "extra_end_insert_rows_rc_func"),
    "add_rows": ("MT", "extra_end_insert_rows_rc_func"),
    # resize
    "row_height_resize": ("RI", "row_height_resize_func"),
    "column_width_resize": ("CH", "column_width_resize_func"),
    # selection
    "cell_select": ("MT", "selection_binding_func"),
    "select_all": ("MT", "select_all_binding_func"),
    "all_select": ("MT", "select_all_binding_func"),
    "row_select": ("RI", "selection_binding_func"),
    "column_select": ("CH", "selection_binding_func"),
    "col_select": ("CH", "selection_binding_func"),
    "drag_select_cells": ("MT", "drag_selection_binding_func"),
    "drag_select_rows": ("RI", "drag_selection_binding_func"),
    "drag_select_columns": ("CH", "drag_selection_binding_func"),
    "shift_cell_select": ("MT", "shift_selection_binding_func"),
    "shift_row_select": ("RI", "shift_selection_binding_func"),
    "shift_column_select": ("CH", "shift_selection_binding_func"),
    "ctrl_cell_select": ("MT", "ctrl_selection_binding_func"),
    "ctrl_row_select": ("RI", "ctrl_selection_binding_func"),
    "ctrl_column_select": ("CH", "ctrl_selection_binding_func"),
    "deselect": ("MT", "deselection_binding_func"),
}

MODIFIED_BINDINGS = [
    ("MT", "extra_end_replace_all_func"),
    ("MT", "extra_end_sort_cells_func"),
    ("CH", "ch_extra_end_sort_rows_func"),
    ("RI", "ri_extra_end_sort_cols_func"),
    ("MT", "extra_end_ctrl_c_func"),
    ("MT", "extra_end_ctrl_x_func"),
    ("MT", "extra_end_ctrl_v_func"),
    ("MT", "extra_end_ctrl_z_func"),
    ("MT", "extra_end_delete_key_func"),
    ("RI", "ri_extra_end_drag_drop_func"),
    ("CH", "ch_extra_end_drag_drop_func"),
    ("MT", "extra_end_del_rows_rc_func"),
    ("MT", "extra_end_del_cols_rc_func"),
    ("MT", "extra_end_insert_cols_rc_func"),
    ("MT", "extra_end_insert_rows_rc_func"),
    ("MT", "extra_end_edit_cell_func"),
    ("CH", "extra_end_edit_cell_func"),
    ("RI", "extra_end_edit_cell_func"),
]

visited = set()
ALL_BINDINGS = []
SELECT_BINDINGS = []
for o, b in BINDING_TO_ATTR.values():
    if b not in visited:
        ALL_BINDINGS.append((o, b))
        visited.add(b)
    if "select" in b and (o, b) not in visited:
        SELECT_BINDINGS.append((o, b))
        visited.add((o, b))
visited = set()

ICON_CUT = (
    """iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAAAXNSR0IArs4c6QAAAxxJREFUSEuFVstq"""
    """20AUPVcqBLLqoiZxVvK2u1JKoFCIocs0/5BlIP6AkWyJOI30AQ5k6X5Dkl1psiwhP6FV3JKuA1lYt7nz"""
    """sEaW3Qps7NHoPs4594yIADDal163XwRe+Vs/aQOsjkOgdTfqtOtKaBbV2rUoTvbZu+Oi+IAKIwD79vFr"""
    """BHyaqvj+35WuT7boQIJThTvTNFcEAQaBRirA7ihW9x4idURbXFYUUQhMZU8Wq76BXaLYDsZfi0sAX5h5"""
    """+uf37EiW32x3Lwh0COAqTdTBIqqPx0s5WV5ErxhTrrAHwrcsUYeOVz9BJdU+/nrYnEzOn4WDwfFgo9Pd"""
    """eSKARkMVCtfLpevKmafEtMeMMh2q3lJ+88j4LK8A4sfZw+bk3CYYDDY62ztP0muaqLApOEJW5FHIGpY9"""
    """AsrRUPWs3OpmXbbxmQ/Rw5Hg19nufgfw6eXzI03izwZT04YJTjo4gDJLVM/Iuk246cAo6M5oiqUbWRaS"""
    """5ZrNA3w8UaqUP4J5CJ6CSQdPk7jnR27MUJ2UMC7yJZnyDUBvCegKvvMQfUngYDHBVW+h8zUDu+q+I5+z"""
    """PI/Cim7BiEBcgqiEVguXWRL3ajk2FNCY8P+MpOBdRGGFW4AjzYJWS9yrLaRZY0NFNV5+nppMjbmWoqcW"""
    """Dcu6q2lOBopmbOHLrbME94bIqMUmHStVtg1xqRu/g7YX8Q2IXoPxHnaI6snPxVXKeYD+iYrLlfqUHS7B"""
    """OPe9CJVeZwQgqZdmaaJ2XKM+ZKIkk0SVzhBdEc673CRfAtTwos7Wzk8Q3rW8yPEixGt1eUlWIG4TFMaL"""
    """tFVMnmXRedFqq2iqS8NF6J/EcemmXWLUdu28SJvd5FkgOrZepM1Oe5Hg1fQCI2G+BShycGnyLU4LFTW9"""
    """aHYkGzpb3QsiOiTgajRUB2JkjerssaqtQ88JIoCtdej6fZLFJsh6kSXZHjgIsJsqdW8ecWe1L0frrBUy"""
    """2ZMmSs4Q00NLpnOMQLxvT7RrBDh1wRf8rTumW1PrTrQ1JtVAu/X24C+0brr6m28VvrO2sF6QVr/C+K8t"""
    """vvbr34S/2B2/MddUJi4AAAAASUVORK5CYII="""
)


ICON_EDIT = (
    """iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAAAXNSR0IArs4c6QAAAn5JREFUSEuFVu1h"""
    """qzAMPJEMQjYhk7x2DEIDoR9hDPomSTd5HiRYr5INCHBa/yK29XGnkxyCLALA+jWtzVbcSFxNO4ie5L75"""
    """XEWxJ2nPeqN9e79k2a5hhgPha8jQtmXpJHNSO2v8Y6YWLKHprvmO0YO5CCeyJElyw4Dnti6/DIJo/Iib"""
    """yCWB1UXTdfme0TOjAODuGY5imt39E2VZI0jqU6l7CZIe8yEnddfFzIPzuioPoxMJvPP4N+6HABtaEpzF"""
    """i0oL0MOrc1luyPjYlicnP5q3a7HL6CYw65eSQg02UQhKxapAmp1yHjKPAXKA3T2jIzGgwRkFD75t6upi"""
    """AliytsSNzkmMGa5+KQ9C1d7jBnAuhQXBiXP6RnWOtG0DxKwtgpkWVYs6F9haj6uiugHIo4pcXZ1iTYSJ"""
    """5JqJm2gRzokXxkLumrbgfO6nHxutkezAPZiiWk6HUaaS/6oPJmTrkiYxKL9zE01SHLEF5xTURFaqUSBT"""
    """3yTcr2E3VXmwSgvnEVks+Jz1soc2jRaaiHtiKjg2kTVR57EPVC2x4JYWG2zRaHPmXND3PAlSmyH/hizF"""
    """9WLYvV67W+CUXVOdIi3BbOHc0mJn1GZyIk5THbkfFx1Sq9mSVMsKmUW5RjHV4PWjkwGVDx7H9hzG7Dg1"""
    """7Xiwgy2R8ObxsQFUKHVVTnuhD+bZI84fjUE7k22vqLPX9+4JxD1An3VVPmsg6YN5asY+2EpwMShtPcZv"""
    """MWmluIzCe9/K71iLMSmnffD4iVBabNZK7vgEaIBO1KPStG+/Y+//NufqspHfhqflxoIiDSAUZfwHTLk4"""
    """9Xv6HB+QxGs0xUvHsWjGlzoth1//zjwys538H8mvnjDSt4tXAAAAAElFTkSuQmCC"""
)


ICON_SORT_DESC = (
    """iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAAAXNSR0IArs4c6QAAAf5JREFUSEuNVW12"""
    """gjAQnMWDVG+CJ+nzGJEWEIsco97EHoWDKGmzhLD5svL8ISG7szszmxCSD4GgodMfE6sEiN3yjdYXm5QX"""
    """fAC5RyZakXwAWYEHwMF2bzqpV9vzLl2hBi6ouhmG7UbjGxrbR4F9p9S40JWuVXZv6xeFuqrNpyU5aZQ2"""
    """6Xi3IK71iMYnFEm+ROUlrxsE4t94J+xPRzXyekKnSHmvA1E5JpQgmERbG7T8Hw1dJ2VB4NNiMaWMrhau"""
    """qOuHGzRM5WNdqV3XD8zQo8BuM+E2A+qxqY67mbq84Embns59S0XxbpKb4K6/cJ66UsTUTWjs+yF2mk+Q"""
    """b33xTSIvHRiAV/wum/J7exUgIvm1aReD5rd57getNVB/zB10F6ORLue0wmLsNGu3JQXhp67Ufj4RhO3k"""
    """MRBSZDViHbxntbI7jQi4flbqwKll5zKQAWwHuemNPJkYhmwHIUVebKjieoRFfogBbHBOg9n/QoOId1nK"""
    """okE0OEAIEGsQCAtc64V3cTtkXOQPWkStXUjr5ysmhshPI0XOAaQHMLis/LHWxrp8o537ixkDPip4Dobh"""
    """himYAyLrd9mSPKVI2lScr38o7fmrLah4qz+OBzNabde3tCka3rXS73hPONS5OHFN5ipKkJW/a5x1XFRu"""
    """7z85AlSpwUJRwqZPBUwMWno/8Au5DCkvTFG1swAAAABJRU5ErkJggg=="""
)


ICON_SORT_ASC = (
    """iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAAAXNSR0IArs4c6QAAAhZJREFUSEuVVu11"""
    """wjAMPAGDhE3KKIyRBBIaXuIxYBPYpCwCbmU7sfyV98qfFlv2Sbo7GcLyIQA6/GaWwnUT4Jb8DoGg3Wm5"""
    """yesxQPk+f3cmRgLLO1MAv7tkmj8sA8P/ZeECwAW53WGaHrzSNfWB/w5KPfDBl0yeCM9zWx+Wzrqzpl2i"""
    """Uy5T2UfgOirNjHRtTRx7uY4X2lBvg+2HgPv5VB8FddmOZ2olXMdpAUi6FpQhhJBwE5Ac4vgKGjLqKhGb"""
    """TW/W4z8Ahkk9oCUHGqap3EfiBOg58xXKtyCGYVTGFGscREfvneEj7F9BRYAB0EB3ci0qaTVjxpiibBeH"""
    """cQoq8B7PObnkiZkDqV+XqVeRrWDmwGybvkveHR9ev46PFZIlB3zuexgvtN30Zf8GFrif2/rovOJh5cDK"""
    """+6CsVakcKWvj0rm/kq/YyQEHCeGxcubJyup1AHb2EN7A8btpXpLkXqlqq3HjVLqmORTHdYZLK1MysjRG"""
    """AvB6b/Rh96Efzvq9wd5c/sEXSN/7tjmm1STDJHgzzJd+UtX2AwapQHgBqCy4fkKTAe7aep9MMz/4Yo9Z"""
    """KiQ5vZosCHRlZ6vVJIFe51O9j6dmxmOJyAIAvrLjSjSDuAqCzOVzuGYu/3ymo+IvZyZ1p3HTGhW3JX6h"""
    """bGUhWCLTuXXJ4UySPqa0WZ7Zi4pCM2RUUe6I/DESRa2MipSt9Zcrsuty/BfWPCAzd/TdsgAAAABJRU5E"""
    """rkJggg=="""
)


ICON_PASTE = (
    """iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAAAXNSR0IArs4c6QAAAlFJREFUSEt9Ve2h"""
    """oyAQnI2NmE5MJe9SBppINKeWkXeVJJ08CjnlDgRcRPWXIDjsfCyE8BAAvQzX83ufkx18IYGWoXtj3+Vz"""
    """KLITXgTkEbQZkAaB1N8J1+YuPhYn/rc9r5lyDwcgEDSafnhD6yJaltao6kqcN0oPmAHZQDye3YNOpy8A"""
    """ud8kK0Fb5LXdwKfV/xqauiq/OdVRBc+u/6U1XqEuR0VdlaYkQMdCPLtBb1B3nUFmzmKAvn9rjWKccDG8"""
    """tpYiFHvl2/kTPrUQFzkMeTbhB4CSlTh74Aig7Xo7X1fCil8bkTPIAxA1WpHLj6Gl/T0Y5e3+zQo4wOGp"""
    """dz623WAqyBeAlT28aL4CTzmnf8OJAc7Q1AhhxI6MExYkALGBV0afP1qdoDESXZtSKGMEY/GgAT/RAlBS"""
    """SAlXywVnnpp3MiOokXAxIFGVfOBtFyg6rGDeKfshzzTeLjdqPOHSCqH8IZiLAFuBBuqbqyARM86BH8mh"""
    """z7OJjKVzECyI12IGcCv3Kghc+0zYAB76TNU3cV71ImABKMkKxQ7czC1EHv7WN0Giz70Sl1nw8BDWQeNi"""
    """+uAsfWZpjrLv80zTywVyI8luLbcpP+mSg7il27SbFqH1izQVGkg6K7sP9imKpY3jMNvUtvSI9w0XxRRx"""
    """6pJMMPv69p7cCazY8C8ucsL1NmcBajHFKsncpiyV17oS396+CUVJAAF3vb71NDXyXj02u2ly4dhVzvR7"""
    """3l/NJwD8dO7SKED4Wu6ADYAEzE4oPU5/ZH17cHqP87juu6vGlxgh0Qn4B4YcZzGhG9dFAAAAAElFTkSu"""
    """QmCC"""
)


ICON_CLEAR = (
    """iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAAAXNSR0IArs4c6QAAAjRJREFUSEuNVotx"""
    """wjAMleC6R9gkTEIzBiF2yAcYo3QS2KSeozRxT/4kcuIAuYMzlqPPe3oyCOxBANDm97iiNYJ2+6GFvzt7"""
    """zRnJE1ta98+eMPQkIW70Xvle1ZyOiLgDgORZRnrmaKhSQd9XUhZXnrZZ1/XpE1b4FWQfSTdeQQg09n0m"""
    """5MEEQf9C3Z5vAJB2uttWQtyXoOM8jdzY3bK8JOsP+EHQShT5xrNpmKubsyFAFjlOqYxAO6FpPFG3F03B"""
    """rB/ryULU8gBLNL8ONQQQOYJGDlEYOd6CYcvGGmFWgckpApHHl7ihM38PzKpqr2gtDdbaNIQs8u0A6cxP"""
    """SDLjYMy/bi83AJ0CgOoeuCWLdY60d5Vinw0qdFBTEsJzwLooGsB2hqYOI20o80FIQYOS1CkTWkIOYiQb"""
    """csK+LsszD0JG63z6cIg8yZ6ounEkO4MfAmR3VRDmBAulpbpf3BInXLGUF1VAGYogQKRNfeWhc1QucAKA"""
    """qntAEGSupyckj9yMJBMsIyeYAGglRb55k+S4DmgArhB3gjlyQUrCShb7jCv/TSW7/L1IKEW35ndDXGjh"""
    """RLCjIiKQsI+sGN2XM0VUvcDlMIu4QOLT6MVNZxLlXTQZ1wiQ6l5n9sLwYliaP/PBVzZNusY1ifJuRwgX"""
    """WuzCmeG0NGUn+5Rkebh66ozVdMa6T2FFVybNGU/sADnX3ugxRE3pXn+X4nDkAhwO89sq7m2pJAfjrBEY"""
    """RGGRr29ej+/i/xAX7B9y8mgsRjNIxAAAAABJRU5ErkJggg=="""
)

ICON_DEL = (
    """iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAAAXNSR0IArs4c6QAAAiJJREFUSEuNVdFh"""
    """ozAMlWCQdhQySZsxDIkJpuAxLpvQUTpIG9/ZFljGojl/JIAlPz1JT0bYFgKCA4cA9JO2xKdgCEB/R8Z+"""
    """+1cjM9n3fyf1APCaH+K+EHC4durOYgyYfKEcQGQzfNh3QPcnvsXF7f3z9wNOw1V9ysS95+bFKROAtQs8"""
    """oPmhQzhAP9qmrmABhE/dqtMeeE11BBBiCwwmGwLXnSI7qhPZm2mm/RZTbnKojYGZ7AIOmgjm/cgHhRLR"""
    """9mEXEKuVcThhMNOtqqp+V6PijCdNs9rf9UWdQ5MlB8p7kZasJ2IrC0k13s8B6Aulk5U0i9J8WAfoQHc+"""
    """r2wVVcxT9x8A8YRguBZWOPQo778DMDpS55h59u0IWqmTBzCzXUIQbXz3a5ys85FF5qmSmdA8TgRw0Hct"""
    """rmbcObVvnsZko3Z15erx0Qkp4t9SEKQPknDSBAfIuihSzQBIUOM0E/3oPIRG4ALM/fLOZAzk6Lhzi7LC"""
    """c5s0ZbNZdMxgPxJKlrzIlKJMB2wUSjoIBWQiijYAfae2RpBAQ3L20o+GeYcM43TDqnrRnTp7+xt7j7wR"""
    """eJGTVtYUMfWskaTo8purmEW8iwLLfLIW94FEX1ZvLnO5TXdFDl3kVepcA4Bnza9DYcCtEP04N3WFdPm0"""
    """pzQQBQBtbFPXbpHvA34RiM9FUMWo0Na+1g9oEOHNbRfQ0Yjbvn958vqi7vtZXt7JT896XhFu8RczqWgy"""
    """PKwKugAAAABJRU5ErkJggg=="""
)


ICON_ADD = (
    """iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAAAXNSR0IArs4c6QAAAYZJREFUSEuVVdt1"""
    """wjAMldNBWKUdhTEC1IVAyBpsQjdLjx+JJesqTvPB4ThGV/ch4RwRzfGDdjz5Ir/PC4A64Qg+qka8JSuo"""
    """5jh+fqkB/kcpN6cqr027Uk/27GiWqgFgiwEnmwC4SBYDRzSM0ztc9af+SwtWKchIrdJisPw6ADym2Is/"""
    """96Z3RZvUefsiM3YYn/En/tIHBfUD0rUjReVXw7iHgTRce2A1FiXKDLJE2K46LMYUxBRVpgQG4eg7AoCZ"""
    """UGFZPNhIDtdaAshqPNa1FYLD8JzeNM+fjrCP8XIwWLiXDxz9hghvAtzu49V13Q9gW45iPY6Sv8/08pfT"""
    """kU8W68PWdHljS2RtDEfGJNcWpgLrHPBBU9MpU5MY7DZZxlTsHLZyhAe4gWTxElMuUVkV+Y5Up7Sfm2YM"""
    """gCxgDvAuwg2FiiJsrclsmmxsATjL6PB2f1xd93Hw5/4I1zU3hUlXb9hNwNZKr7esngO1ci28Rj4RAzy9"""
    """Mi0Yrt6g5e+WDRroSLHhYPsY/AFsVMwga5Ld0wAAAABJRU5ErkJggg=="""
)


ICON_COPY = (
    """iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAAAXNSR0IArs4c6QAAAjdJREFUSEuFVuuZ"""
    """qjAQPYOFXLcTrGTXMiIIImLK0FuJdrI2smRvJgEmJLn6w4+Qx8x5zATC/CMQDAwB/m+ZEk9ymp8Tq+R7"""
    """mgZdP5yoKD4BbOc9drddsA7gxq9/s11TqXsyE35J4ADdRX+BcEtnJKMYu2Wd9T4ZxB/G+Z2v+gGD8mfE"""
    """rmvU838UTdm2+rrdjPQNg1dTq480CpuODTBoprKp1ExIwOPMdYiRaaXiT1OrvdRSKpMNkOGdM3onbLDX"""
    """a2DJDRB4jVaneae9cU7oTEvRRRuQQVMdVp4Buv5yomLzCYOtXcO2SrhLxAzc5SjiAB6B4ODM7jK3hHOW"""
    """80KTwSbB6w32Ta3uIkCMwLnLlD8j7bqjeq6ZSdVkq/V2M+Lb1klTqY+pDiKKWJucu96U8+LKA7kAiYMW"""
    """+3pkkX3SRrb7+kEby5zVNNZA8BBok+8H0YysK4EgoYFEFhEeQlpGhPNwnQvXIdD6gRElAftjre5Tq3CZ"""
    """JOwr+szUfbkTeywpDb4A3HxxPZuD2qW1yfEeouEABmhq5TRw1jIliGxBWXtxb5kKsK0O5DLNNZBQhkjk"""
    """qcGsz0hRlPfOFITQD1fvIo9gmQqTjLus7EX5UJGLHPK4T04IIoq8o6SwkiSmiDXwdRAwKBzS+YsIcH3F"""
    """uit96JJc2+tyU+ABgjNLrr97kfkqTdeYvD7nBsdE+DP5KnUBMnQG7hpRzh8AuXbt3r/MOP5tj9VpvvQn"""
    """98lqZLmjwPLTRk7Lygvx/gLRN1wxA1X+ogAAAABJRU5ErkJggg=="""
)


ICON_SELECT_ALL = (
    """iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAAAXNSR0IArs4c6QAAAiZJREFUSEuNVe11"""
    """gzAMPJlBmm4Ck6QZwyF8tsAYTSdJNmkG6cONhQk2tkn4kUeELEun04lgPQRA8Y8xGgNBGZP56Dna/nZE"""
    """QLsGnlgg7/bH2eULYUlodUH92dVCiD0It+ooM11I2w8XKKQgXMujzHSghm0qBdG1zGWmy2O/ETsQmjKX"""
    """Z05dWRe0Xf8B0PeUkjqX+fGg35rPriYhKgDnMpe+7SQPOhD7kagMJge+5H7+AZHOgBTSvxFZU8hrGLoY"""
    """mlO61TDskhG/AG7lSb47FZhM3+YsYxeEu7DgruMIId4Krkzb5zr4z8yWV/pu+9ukmGCZiEggO6NNRgWY"""
    """6J0NBHN6oGPMrFgq83ntzMrq4rbvLzpzzTi3yV3PIBW5JC7tKVwBHwLar4Hvn+LAQHQP2bRMR7fJmxMb"""
    """xmMdx++BHdR7txsYgC7CjbhYxJVhdSbGonnQNHad3RyDb4Am3BtPBJfKdJNNL1lCFhbNTT5JWkvq84Yv"""
    """BbXdwCNQOk1mzfmqSSSmyc+nI9aNtSI4cr0J+StTZQ/PDKPHxBg1vcOB0bZNnlxrPdfY8QTGIdoEj8li"""
    """x1mzSFEKgLU8hogtZMtunQKV7ZAmibrwIvKlYvgAMC0cs730q7vRjplmVDPw9kqJcC1MoIff/TuBDsVJ"""
    """nhe5vstrNfS7ZCS9GvdQuPFeYAlhdr220YTYq3H8qYq8nokblWtH063hWsMyD51td/iwSZoYUV6WEOAf"""
    """jjY7LKuzEksAAAAASUVORK5CYII="""
)


ICON_UNDO = (
    """iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAAAXNSR0IArs4c6QAAAjpJREFUSEuNVv19"""
    """wiAQPdBB4iZmk3YMYg0h/jRjtJ1EN5E92kB7BMIRiJp/QHLc53svMsCHAYB1u8KevvQ2ryz+Gi7zU3a1"""
    """Fj1NKFqlXpIAz6vxBZLsrNszYGDTJjyuYLrQDkO1sfCJgdtG1MvOvNK8vEX+FjrfAnyCgT0AfLUH8V4c"""
    """1pMozFfohhxsZcgcnTPQ7UHsXplryYZUMLmfnVuXedn5q73BpGkF1DkD0EefObWhQHiOuhDA13a6DFfr"""
    """MkdSML/4Pa1/OtL+6DZyUL0QOlBpmcR8NQZIvU2YwqcQbDLVxpjv7njo0jngTcJe1yID13/UVNiiXw61"""
    """Eo2OCIhNQVu8yn/NG+Nc4htjjJKLIBnRaBDMbORQu/IfDFYOl2pj2B0s6NHad3VsbgGSSQVhmPLiK7FQ"""
    """Ya8xiGqEplBeQrI7nTvOuAQON9mIemppaNH8wyLrXUdiJbYCYLr9ELtZCzLAM3BVjHDHhORB7Kh2pqpF"""
    """Lvsg0kmFYzLlctozb3tPubOoIGr2EgtOzxK2zxah4tOw3zC4zi0K9msyW1bWXIpDYHU6d4imqFt+Bhk4"""
    """6AGV4uw8Srfsh/2WA5IURgt1fxS3dAYFR2uBacU4WP5j3tjGZQ52NEq2H138NiQzSD8mcQqTy/4yXMHa"""
    """vWO1I3TKajuTLG/jDI1sHsRW9ecOM3VH0be2xnybLf9SQgRtmjgWupIp5Rpo15hcmBn9BzHJdYE4DyGb"""
    """QSzlh5tBDtO1GdDoBZgutXmR7R8TrUAvyr2/1AAAAABJRU5ErkJggg=="""
)


ICON_REDO = (
    """iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAAAXNSR0IArs4c6QAAAkJJREFUSEuVVe2B"""
    """oyAQncErRCu5tZJNykAjijlDGclVEjuJhZywgqggmN3zT3QY5uPNmxeE/YMAoKzRvpuft3YEBOX7OCHW"""
    """1yOnrYajjG5BbiUA+iv6RAqG2X1pL3ItcjwnCKIdeDqx25t46qsjwplTOgRdxiAKwD8aEAK03e0JCj8A"""
    """YBgJ5FuSH0J0BN1ir4VIEwm6i3SfZA/If0Br2WKzBEkQcl4UgyGLB9Hb2blBw9HXN5EmKt7JMt4oGro6"""
    """8k+eMCG/QYHG2jJoIZ6ucXs3ZRiTGlhZZDNxEDBYosmHX7sGCam9zDae8Z8DOQkcT4SeFTRfLM4ezFA0"""
    """JjjW+ktKySWQvq1o7y23g5SZg1J3wygFA7sUmWGjs5NreiZE+kvCSzuMEnPOaG/aXGXAH9YcHO4aQpzo"""
    """WpU0iynPart24q4UnJSSvK7KJlzA7Xotbmmi0AQHUENdFpmnVzEWtZ14aV6PBLKW0uFQFMyiiacd/sBK"""
    """msVmGbCo7YQhMLtQT6NiusSvfxokyacOvqpNTHG3zAjtTa8/fIxS5W1V9EvLbnXxSl2V9afgVdp24j6t"""
    """/WmC6cFKel64HChoEO+7BPbcsGKEl16WaUV4VdKH334oZLHBbgqtGbh7eNs1mJglG9Qo/0okPa9o7/L6"""
    """MKjDHF+LbErNd70LWiIIIfUhiwj0jNI8LmG+IAYdLA3ZpasBITV09KXnwS707EOxKehbiL77H9jOY+Sd"""
    """T/esW+8EFIwYXK0Pki0K6khLCFEAbBzpn3b6BQruMyxajTd+AAAAAElFTkSuQmCC"""
)



import tkinter as tk
from tkinter import TclError

from .functions import get_menu_kwargs


def menu_add_command(menu: tk.Menu, **kwargs) -> None:
    if "label" not in kwargs:
        return
    try:
        index = menu.index(kwargs["label"])
        menu.delete(index)
    except TclError:
        pass
    menu.add_command(**kwargs)


def build_table_rc_menu(MT, popup_menu: tk.Menu, r: int, c: int) -> None:
    datarn, datacn = MT.datarn(r), MT.datacn(c)
    popup_menu.delete(0, "end")
    mnkwgs = get_menu_kwargs(MT.PAR.ops)
    if MT.table_edit_cell_enabled() and not MT.is_readonly(datarn, datacn):
        menu_add_command(
            popup_menu,
            label=MT.PAR.ops.edit_cell_label,
            command=lambda: MT.open_cell(event="rc"),
            image=MT.PAR.ops.edit_cell_image,
            compound=MT.PAR.ops.edit_cell_compound,
            **mnkwgs,
        )
    if MT.cut_enabled and any(x in MT.enabled_bindings_menu_entries for x in ("all", "cut", "edit_bindings", "edit")):
        menu_add_command(
            popup_menu,
            label=MT.PAR.ops.cut_label,
            accelerator=MT.PAR.ops.cut_accelerator,
            command=MT.ctrl_x,
            image=MT.PAR.ops.cut_image,
            compound=MT.PAR.ops.cut_compound,
            **mnkwgs,
        )
    if MT.copy_enabled and any(x in MT.enabled_bindings_menu_entries for x in ("all", "copy", "edit_bindings", "edit")):
        menu_add_command(
            popup_menu,
            label=MT.PAR.ops.copy_label,
            accelerator=MT.PAR.ops.copy_accelerator,
            command=MT.ctrl_c,
            image=MT.PAR.ops.copy_image,
            compound=MT.PAR.ops.copy_compound,
            **mnkwgs,
        )
    if MT.paste_enabled and any(
        x in MT.enabled_bindings_menu_entries for x in ("all", "paste", "edit_bindings", "edit")
    ):
        menu_add_command(
            popup_menu,
            label=MT.PAR.ops.paste_label,
            accelerator=MT.PAR.ops.paste_accelerator,
            command=MT.ctrl_v,
            image=MT.PAR.ops.paste_image,
            compound=MT.PAR.ops.paste_compound,
            **mnkwgs,
        )
    if MT.delete_key_enabled and any(
        x in MT.enabled_bindings_menu_entries for x in ("all", "delete", "edit_bindings", "edit")
    ):
        menu_add_command(
            popup_menu,
            label=MT.PAR.ops.delete_label,
            accelerator=MT.PAR.ops.delete_accelerator,
            command=MT.delete_key,
            image=MT.PAR.ops.delete_image,
            compound=MT.PAR.ops.delete_compound,
            **mnkwgs,
        )
    if MT.rc_sort_cells_enabled:
        menu_add_command(
            popup_menu,
            label=MT.PAR.ops.sort_cells_label,
            accelerator=MT.PAR.ops.sort_cells_accelerator,
            command=MT.sort_boxes,
            image=MT.PAR.ops.sort_cells_image,
            compound=MT.PAR.ops.sort_cells_compound,
            **mnkwgs,
        )
        menu_add_command(
            popup_menu,
            label=MT.PAR.ops.sort_cells_reverse_label,
            accelerator=MT.PAR.ops.sort_cells_reverse_accelerator,
            command=lambda: MT.sort_boxes(reverse=True),
            image=MT.PAR.ops.sort_cells_reverse_image,
            compound=MT.PAR.ops.sort_cells_reverse_compound,
            **mnkwgs,
        )
        menu_add_command(
            popup_menu,
            label=MT.PAR.ops.sort_cells_x_label,
            accelerator=MT.PAR.ops.sort_cells_x_accelerator,
            command=lambda: MT.sort_boxes(row_wise=True),
            image=MT.PAR.ops.sort_cells_x_image,
            compound=MT.PAR.ops.sort_cells_x_compound,
            **mnkwgs,
        )
        menu_add_command(
            popup_menu,
            label=MT.PAR.ops.sort_cells_x_reverse_label,
            accelerator=MT.PAR.ops.sort_cells_x_reverse_accelerator,
            command=lambda: MT.sort_boxes(reverse=True, row_wise=True),
            image=MT.PAR.ops.sort_cells_x_reverse_image,
            compound=MT.PAR.ops.sort_cells_x_reverse_compound,
            **mnkwgs,
        )
    if MT.undo_enabled and any(
        x in MT.enabled_bindings_menu_entries for x in ("all", "undo", "redo", "edit_bindings", "edit")
    ):
        if MT.undo_stack:
            menu_add_command(
                popup_menu,
                label=MT.PAR.ops.undo_label,
                accelerator=MT.PAR.ops.undo_accelerator,
                command=MT.undo,
                image=MT.PAR.ops.undo_image,
                compound=MT.PAR.ops.undo_compound,
                **mnkwgs,
            )
        if MT.redo_stack:
            menu_add_command(
                popup_menu,
                label=MT.PAR.ops.redo_label,
                accelerator=MT.PAR.ops.redo_accelerator,
                command=MT.redo,
                image=MT.PAR.ops.redo_image,
                compound=MT.PAR.ops.redo_compound,
                **mnkwgs,
            )
    for label, kws in MT.extra_table_rc_menu_funcs.items():
        menu_add_command(popup_menu, label=label, **{**mnkwgs, **kws})


def build_index_rc_menu(MT, popup_menu: tk.Menu, r: int) -> None:
    datarn = MT.datarn(r)
    popup_menu.delete(0, "end")
    mnkwgs = get_menu_kwargs(MT.PAR.ops)
    if MT.index_edit_cell_enabled() and not MT.RI.is_readonly(datarn):
        menu_add_command(
            popup_menu,
            label=MT.PAR.ops.edit_index_label,
            command=lambda: MT.RI.open_cell(event="rc"),
            image=MT.PAR.ops.edit_index_image,
            compound=MT.PAR.ops.edit_index_compound,
            **mnkwgs,
        )
    if MT.cut_enabled and any(x in MT.enabled_bindings_menu_entries for x in ("all", "cut", "edit_bindings", "edit")):
        menu_add_command(
            popup_menu,
            label=MT.PAR.ops.cut_contents_label,
            accelerator=MT.PAR.ops.cut_contents_accelerator,
            command=MT.ctrl_x,
            image=MT.PAR.ops.cut_contents_image,
            compound=MT.PAR.ops.cut_contents_compound,
            **mnkwgs,
        )
    if MT.copy_enabled and any(x in MT.enabled_bindings_menu_entries for x in ("all", "copy", "edit_bindings", "edit")):
        menu_add_command(
            popup_menu,
            label=MT.PAR.ops.copy_contents_label,
            accelerator=MT.PAR.ops.copy_contents_accelerator,
            command=MT.ctrl_c,
            image=MT.PAR.ops.copy_contents_image,
            compound=MT.PAR.ops.copy_contents_compound,
            **mnkwgs,
        )
    if MT.paste_enabled and any(
        x in MT.enabled_bindings_menu_entries for x in ("all", "paste", "edit_bindings", "edit")
    ):
        menu_add_command(
            popup_menu,
            label=MT.PAR.ops.paste_label,
            accelerator=MT.PAR.ops.paste_accelerator,
            command=MT.ctrl_v,
            image=MT.PAR.ops.paste_image,
            compound=MT.PAR.ops.paste_compound,
            **mnkwgs,
        )
    if MT.delete_key_enabled and any(
        x in MT.enabled_bindings_menu_entries for x in ("all", "delete", "edit_bindings", "edit")
    ):
        menu_add_command(
            popup_menu,
            label=MT.PAR.ops.clear_contents_label,
            accelerator=MT.PAR.ops.clear_contents_accelerator,
            command=MT.delete_key,
            image=MT.PAR.ops.clear_contents_image,
            compound=MT.PAR.ops.clear_contents_compound,
            **mnkwgs,
        )
    if MT.rc_delete_row_enabled:
        menu_add_command(
            popup_menu,
            label=MT.PAR.ops.delete_rows_label,
            command=MT.delete_rows,
            image=MT.PAR.ops.delete_rows_image,
            compound=MT.PAR.ops.delete_rows_compound,
            **mnkwgs,
        )
    if MT.rc_insert_row_enabled:
        menu_add_command(
            popup_menu,
            label=MT.PAR.ops.insert_rows_above_label,
            command=lambda: MT.rc_add_rows("above"),
            image=MT.PAR.ops.insert_rows_above_image,
            compound=MT.PAR.ops.insert_rows_above_compound,
            **mnkwgs,
        )
        menu_add_command(
            popup_menu,
            label=MT.PAR.ops.insert_rows_below_label,
            command=lambda: MT.rc_add_rows("below"),
            image=MT.PAR.ops.insert_rows_below_image,
            compound=MT.PAR.ops.insert_rows_below_compound,
            **mnkwgs,
        )
    if MT.rc_sort_row_enabled:
        menu_add_command(
            popup_menu,
            label=MT.PAR.ops.sort_row_label,
            accelerator=MT.PAR.ops.sort_row_accelerator,
            command=MT.RI._sort_rows,
            image=MT.PAR.ops.sort_row_image,
            compound=MT.PAR.ops.sort_row_compound,
            **mnkwgs,
        )
        menu_add_command(
            popup_menu,
            label=MT.PAR.ops.sort_row_reverse_label,
            accelerator=MT.PAR.ops.sort_row_reverse_accelerator,
            command=lambda: MT.RI._sort_rows(reverse=True),
            image=MT.PAR.ops.sort_row_reverse_image,
            compound=MT.PAR.ops.sort_row_reverse_compound,
            **mnkwgs,
        )
    if MT.rc_sort_columns_enabled:
        menu_add_command(
            popup_menu,
            label=MT.PAR.ops.sort_columns_label,
            accelerator=MT.PAR.ops.sort_columns_accelerator,
            command=MT.RI._sort_columns_by_row,
            image=MT.PAR.ops.sort_columns_image,
            compound=MT.PAR.ops.sort_columns_compound,
            **mnkwgs,
        )
        menu_add_command(
            popup_menu,
            label=MT.PAR.ops.sort_columns_reverse_label,
            accelerator=MT.PAR.ops.sort_columns_reverse_accelerator,
            command=lambda: MT.RI._sort_columns_by_row(reverse=True),
            image=MT.PAR.ops.sort_columns_reverse_image,
            compound=MT.PAR.ops.sort_columns_reverse_compound,
            **mnkwgs,
        )
    if MT.undo_enabled and any(
        x in MT.enabled_bindings_menu_entries for x in ("all", "undo", "redo", "edit_bindings", "edit")
    ):
        if MT.undo_stack:
            menu_add_command(
                popup_menu,
                label=MT.PAR.ops.undo_label,
                accelerator=MT.PAR.ops.undo_accelerator,
                command=MT.undo,
                image=MT.PAR.ops.undo_image,
                compound=MT.PAR.ops.undo_compound,
                **mnkwgs,
            )
        if MT.redo_stack:
            menu_add_command(
                popup_menu,
                label=MT.PAR.ops.redo_label,
                accelerator=MT.PAR.ops.redo_accelerator,
                command=MT.redo,
                image=MT.PAR.ops.redo_image,
                compound=MT.PAR.ops.redo_compound,
                **mnkwgs,
            )
    for label, kws in MT.extra_index_rc_menu_funcs.items():
        menu_add_command(popup_menu, label=label, **{**mnkwgs, **kws})


def build_header_rc_menu(MT, popup_menu: tk.Menu, c: int) -> None:
    datacn = MT.datacn(c)
    popup_menu.delete(0, "end")
    mnkwgs = get_menu_kwargs(MT.PAR.ops)
    if MT.header_edit_cell_enabled() and not MT.CH.is_readonly(datacn):
        menu_add_command(
            popup_menu,
            label=MT.PAR.ops.edit_header_label,
            command=lambda: MT.CH.open_cell(event="rc"),
            image=MT.PAR.ops.edit_header_image,
            compound=MT.PAR.ops.edit_header_compound,
            **mnkwgs,
        )
    if MT.cut_enabled and any(x in MT.enabled_bindings_menu_entries for x in ("all", "cut", "edit_bindings", "edit")):
        menu_add_command(
            popup_menu,
            label=MT.PAR.ops.cut_contents_label,
            accelerator=MT.PAR.ops.cut_contents_accelerator,
            command=MT.ctrl_x,
            image=MT.PAR.ops.cut_contents_image,
            compound=MT.PAR.ops.cut_contents_compound,
            **mnkwgs,
        )
    if MT.copy_enabled and any(x in MT.enabled_bindings_menu_entries for x in ("all", "copy", "edit_bindings", "edit")):
        menu_add_command(
            popup_menu,
            label=MT.PAR.ops.copy_contents_label,
            accelerator=MT.PAR.ops.copy_contents_accelerator,
            command=MT.ctrl_c,
            image=MT.PAR.ops.copy_contents_image,
            compound=MT.PAR.ops.copy_contents_compound,
            **mnkwgs,
        )
    if MT.paste_enabled and any(
        x in MT.enabled_bindings_menu_entries for x in ("all", "paste", "edit_bindings", "edit")
    ):
        menu_add_command(
            popup_menu,
            label=MT.PAR.ops.paste_label,
            accelerator=MT.PAR.ops.paste_accelerator,
            command=MT.ctrl_v,
            image=MT.PAR.ops.paste_image,
            compound=MT.PAR.ops.paste_compound,
            **mnkwgs,
        )
    if MT.delete_key_enabled and any(
        x in MT.enabled_bindings_menu_entries for x in ("all", "delete", "edit_bindings", "edit")
    ):
        menu_add_command(
            popup_menu,
            label=MT.PAR.ops.clear_contents_label,
            accelerator=MT.PAR.ops.clear_contents_accelerator,
            command=MT.delete_key,
            image=MT.PAR.ops.clear_contents_image,
            compound=MT.PAR.ops.clear_contents_compound,
            **mnkwgs,
        )
    if MT.rc_delete_column_enabled:
        menu_add_command(
            popup_menu,
            label=MT.PAR.ops.delete_columns_label,
            command=MT.delete_columns,
            image=MT.PAR.ops.delete_columns_image,
            compound=MT.PAR.ops.delete_columns_compound,
            **mnkwgs,
        )
    if MT.rc_insert_column_enabled:
        menu_add_command(
            popup_menu,
            label=MT.PAR.ops.insert_columns_left_label,
            command=lambda: MT.rc_add_columns("left"),
            image=MT.PAR.ops.insert_columns_left_image,
            compound=MT.PAR.ops.insert_columns_left_compound,
            **mnkwgs,
        )
        menu_add_command(
            popup_menu,
            label=MT.PAR.ops.insert_columns_right_label,
            command=lambda: MT.rc_add_columns("right"),
            image=MT.PAR.ops.insert_columns_right_image,
            compound=MT.PAR.ops.insert_columns_right_compound,
            **mnkwgs,
        )
    if MT.rc_sort_column_enabled:
        menu_add_command(
            popup_menu,
            label=MT.PAR.ops.sort_column_label,
            accelerator=MT.PAR.ops.sort_column_accelerator,
            command=MT.CH._sort_columns,
            image=MT.PAR.ops.sort_column_image,
            compound=MT.PAR.ops.sort_column_compound,
            **mnkwgs,
        )
        menu_add_command(
            popup_menu,
            label=MT.PAR.ops.sort_column_reverse_label,
            accelerator=MT.PAR.ops.sort_column_reverse_accelerator,
            command=lambda: MT.CH._sort_columns(reverse=True),
            image=MT.PAR.ops.sort_column_reverse_image,
            compound=MT.PAR.ops.sort_column_reverse_compound,
            **mnkwgs,
        )
    if MT.rc_sort_rows_enabled:
        menu_add_command(
            popup_menu,
            label=MT.PAR.ops.sort_rows_label,
            accelerator=MT.PAR.ops.sort_rows_accelerator,
            command=MT.CH._sort_rows_by_column,
            image=MT.PAR.ops.sort_rows_image,
            compound=MT.PAR.ops.sort_rows_compound,
            **mnkwgs,
        )
        menu_add_command(
            popup_menu,
            label=MT.PAR.ops.sort_rows_reverse_label,
            accelerator=MT.PAR.ops.sort_rows_reverse_accelerator,
            command=lambda: MT.CH._sort_rows_by_column(reverse=True),
            image=MT.PAR.ops.sort_rows_reverse_image,
            compound=MT.PAR.ops.sort_rows_reverse_compound,
            **mnkwgs,
        )
    if MT.undo_enabled and any(
        x in MT.enabled_bindings_menu_entries for x in ("all", "undo", "redo", "edit_bindings", "edit")
    ):
        if MT.undo_stack:
            menu_add_command(
                popup_menu,
                label=MT.PAR.ops.undo_label,
                accelerator=MT.PAR.ops.undo_accelerator,
                command=MT.undo,
                image=MT.PAR.ops.undo_image,
                compound=MT.PAR.ops.undo_compound,
                **mnkwgs,
            )
        if MT.redo_stack:
            menu_add_command(
                popup_menu,
                label=MT.PAR.ops.redo_label,
                accelerator=MT.PAR.ops.redo_accelerator,
                command=MT.redo,
                image=MT.PAR.ops.redo_image,
                compound=MT.PAR.ops.redo_compound,
                **mnkwgs,
            )
    for label, kws in MT.extra_header_rc_menu_funcs.items():
        menu_add_command(popup_menu, label=label, **{**mnkwgs, **kws})


def build_empty_rc_menu(MT, popup_menu: tk.Menu) -> None:
    popup_menu.delete(0, "end")
    mnkwgs = get_menu_kwargs(MT.PAR.ops)
    if (
        MT.paste_enabled
        and any(x in MT.enabled_bindings_menu_entries for x in ("all", "paste", "edit_bindings", "edit"))
        and (MT.PAR.ops.paste_can_expand_x or MT.PAR.ops.paste_can_expand_y)
    ):
        menu_add_command(
            popup_menu,
            label=MT.PAR.ops.paste_label,
            accelerator=MT.PAR.ops.paste_accelerator,
            command=MT.ctrl_v,
            image=MT.PAR.ops.paste_image,
            compound=MT.PAR.ops.paste_compound,
            **mnkwgs,
        )
    if MT.rc_insert_column_enabled:
        menu_add_command(
            popup_menu,
            label=MT.PAR.ops.insert_column_label,
            command=lambda: MT.rc_add_columns("left"),
            image=MT.PAR.ops.insert_column_image,
            compound=MT.PAR.ops.insert_column_compound,
            **mnkwgs,
        )
    if MT.rc_insert_row_enabled:
        menu_add_command(
            popup_menu,
            label=MT.PAR.ops.insert_row_label,
            command=lambda: MT.rc_add_rows("below"),
            image=MT.PAR.ops.insert_row_image,
            compound=MT.PAR.ops.insert_row_compound,
            **mnkwgs,
        )
    if MT.undo_enabled and any(
        x in MT.enabled_bindings_menu_entries for x in ("all", "undo", "redo", "edit_bindings", "edit")
    ):
        if MT.undo_stack:
            menu_add_command(
                popup_menu,
                label=MT.PAR.ops.undo_label,
                accelerator=MT.PAR.ops.undo_accelerator,
                command=MT.undo,
                image=MT.PAR.ops.undo_image,
                compound=MT.PAR.ops.undo_compound,
                **mnkwgs,
            )
        if MT.redo_stack:
            menu_add_command(
                popup_menu,
                label=MT.PAR.ops.redo_label,
                accelerator=MT.PAR.ops.redo_accelerator,
                command=MT.redo,
                image=MT.PAR.ops.redo_image,
                compound=MT.PAR.ops.redo_compound,
                **mnkwgs,
            )
    for label, kws in MT.extra_empty_space_rc_menu_funcs.items():
        menu_add_command(popup_menu, label=label, **{**mnkwgs, **kws})



from __future__ import annotations

import tkinter as tk
from collections.abc import Callable
from typing import Any, Literal

from .constants import align_helper, ctrl_key, rc_binding
from .functions import convert_align
from .other_classes import DotDict, FontTuple


class TextEditorTkText(tk.Text):
    def __init__(
        self,
        parent: tk.Misc,
        newline_binding: None | Callable = None,
    ) -> None:
        super().__init__(
            parent,
            spacing1=0,
            spacing2=1,
            spacing3=0,
            bd=0,
            highlightthickness=0,
            undo=True,
            maxundo=30,
        )
        self.parent = parent
        self.newline_bindng = newline_binding
        self.rc_popup_menu = tk.Menu(self, tearoff=0)
        self.bind("<1>", lambda event: self.focus_set())
        self.bind(rc_binding, self.rc)
        self.bind(f"<{ctrl_key}-a>", self.select_all)
        self.bind(f"<{ctrl_key}-A>", self.select_all)
        self.bind("<Delete>", self.delete_key)
        self._orig = self._w + "_orig"
        self.tk.call("rename", self._w, self._orig)
        self.tk.createcommand(self._w, self._proxy)

    def reset(
        self,
        menu_kwargs: dict,
        sheet_ops: dict,
        align: str,
        font: tuple,
        bg: str,
        fg: str,
        select_bg: str,
        select_fg: str,
        state: str,
        text: str = "",
    ) -> None:
        self.config(
            font=font,
            background=bg,
            foreground=fg,
            insertbackground=fg,
            state=state,
            selectbackground=select_bg,
            selectforeground=select_fg,
        )
        self.editor_del_key = sheet_ops.editor_del_key
        self.align = align
        self.rc_popup_menu.delete(0, "end")
        self.rc_popup_menu.add_command(
            label=sheet_ops.select_all_label,
            accelerator=sheet_ops.select_all_accelerator,
            image=sheet_ops.select_all_image,
            compound=sheet_ops.select_all_compound,
            command=self.select_all,
            **menu_kwargs,
        )
        self.rc_popup_menu.add_command(
            label=sheet_ops.cut_label,
            accelerator=sheet_ops.cut_accelerator,
            image=sheet_ops.cut_image,
            compound=sheet_ops.cut_compound,
            command=self.cut,
            **menu_kwargs,
        )
        self.rc_popup_menu.add_command(
            label=sheet_ops.copy_label,
            accelerator=sheet_ops.copy_accelerator,
            image=sheet_ops.copy_image,
            compound=sheet_ops.copy_compound,
            command=self.copy,
            **menu_kwargs,
        )
        self.rc_popup_menu.add_command(
            label=sheet_ops.paste_label,
            accelerator=sheet_ops.paste_accelerator,
            image=sheet_ops.paste_image,
            compound=sheet_ops.paste_compound,
            command=self.paste,
            **menu_kwargs,
        )
        self.rc_popup_menu.add_command(
            label=sheet_ops.undo_label,
            accelerator=sheet_ops.undo_accelerator,
            image=sheet_ops.undo_image,
            compound=sheet_ops.undo_compound,
            command=self.undo,
            **menu_kwargs,
        )
        self.rc_popup_menu.add_command(
            label=sheet_ops.redo_label,
            accelerator=sheet_ops.redo_accelerator,
            image=sheet_ops.redo_image,
            compound=sheet_ops.redo_compound,
            command=self.redo,
            **menu_kwargs,
        )
        self.align = align_helper[convert_align(align)]
        self.delete(1.0, "end")
        self.insert(1.0, text)
        self.yview_moveto(1)
        self.tag_configure("align", justify=self.align)
        self.tag_add("align", 1.0, "end")
        self.edit_reset()

    def _proxy(self, command: Any, *args) -> Any:
        try:
            result = self.tk.call((self._orig, command) + args)
        except Exception:
            return
        if command in (
            "insert",
            "delete",
            "replace",
        ):
            self.tag_add("align", 1.0, "end")
            self.event_generate("<<TextModified>>")
            if (
                args
                and len(args) > 1
                and args[1] != "\n"
                and args != ("1.0", "end")
                and self.yview() != (0.0, 1.0)
                and self.newline_bindng is not None
            ):
                self.newline_bindng(check_lines=False)
        return result

    def rc(self, event: Any) -> None:
        self.focus_set()
        self.rc_popup_menu.tk_popup(event.x_root, event.y_root)

    def delete_key(self, event: Any = None) -> None:
        if self.editor_del_key == "forward":
            return
        elif not self.editor_del_key:
            return "break"
        elif self.editor_del_key == "backward":
            if self.tag_ranges("sel"):
                return
            if self.index("insert") == "1.0":
                return "break"
            self.delete("insert-1c")
            return "break"

    def select_all(self, event: Any = None) -> Literal["break"]:
        self.tag_add(tk.SEL, "1.0", "end-1c")
        self.mark_set(tk.INSERT, "end-1c")
        # self.see(tk.INSERT)
        return "break"

    def cut(self, event: Any = None) -> Literal["break"]:
        self.event_generate(f"<{ctrl_key}-x>")
        self.event_generate("<KeyRelease>")
        return "break"

    def copy(self, event: Any = None) -> Literal["break"]:
        self.event_generate(f"<{ctrl_key}-c>")
        return "break"

    def paste(self, event: Any = None) -> Literal["break"]:
        self.event_generate(f"<{ctrl_key}-v>")
        self.event_generate("<KeyRelease>")
        return "break"

    def undo(self, event: Any = None) -> Literal["break"]:
        self.event_generate(f"<{ctrl_key}-z>")
        self.event_generate("<KeyRelease>")
        return "break"

    def redo(self, event: Any = None) -> Literal["break"]:
        self.event_generate(f"<{ctrl_key}-Shift-z>")
        self.event_generate("<KeyRelease>")
        return "break"


class TextEditor(tk.Frame):
    def __init__(
        self,
        parent: tk.Misc,
        newline_binding: None | Callable = None,
    ) -> None:
        super().__init__(
            parent,
            width=0,
            height=0,
            bd=0,
        )
        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(0, weight=1)
        self.grid_propagate(False)
        self.parent = parent
        self.r = 0
        self.c = 0
        self.tktext = TextEditorTkText(self, newline_binding=newline_binding)
        self.tktext.grid(row=0, column=0, sticky="nswe")

    def get(self) -> str:
        return self.tktext.get("1.0", "end-1c")

    def get_num_lines(self) -> int:
        return int(self.tktext.index("end-1c").split(".")[0])

    def set_text(self, text: str = "") -> None:
        self.tktext.delete(1.0, "end")
        self.tktext.insert(1.0, text)

    def scroll_to_bottom(self) -> None:
        self.tktext.yview_moveto(1)

    def reset(
        self,
        width: int,
        height: int,
        border_color: str,
        show_border: bool,
        menu_kwargs: DotDict,
        sheet_ops: DotDict,
        font: FontTuple,
        bg: str,
        fg: str,
        select_bg: str,
        select_fg: str,
        align: str,
        state: str,
        r: int = 0,
        c: int = 0,
        text: str = "",
    ) -> None:
        self.r = r
        self.c = c
        self.tktext.reset(
            menu_kwargs=menu_kwargs,
            sheet_ops=sheet_ops,
            align=align,
            font=font,
            bg=bg,
            fg=fg,
            select_bg=select_bg,
            select_fg=select_fg,
            state=state,
            text=text,
        )
        self.config(
            width=width,
            height=height,
            background=bg,
            highlightbackground=border_color,
            highlightcolor=border_color,
            highlightthickness=2 if show_border else 0,
        )
